from typing import _GenericAlias  # type: ignore
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import MutableSequence
from typing import Sequence
from typing import Type
from typing import TypeVar
from typing import Union

from .compat import _SpecialGenericAlias
from .compat import get_args
from .compat import get_origin


T = TypeVar("T")


def is_alias(type_: Any) -> bool:
    try:
        return isinstance(type_, _GenericAlias)
    except TypeError:
        return False


def has_annotated_init(type_: Type[T]) -> bool:
    return (
        not is_alias(type_)
        and hasattr(type_, "__init__")
        and hasattr(type_.__init__, "__annotations__")
    )


def is_literal(type_: Any) -> bool:
    if origin := get_origin(type_):
        try:
            return origin._name == "Literal"
        except AttributeError:
            pass
    return False


def is_union(type_: Union) -> bool:
    return get_origin(type_) is Union


def is_sequence(type_: Any) -> bool:
    try:
        return issubclass(get_origin(type_), Sequence)  # type: ignore
    except TypeError:
        return False


def is_list(type_: Any) -> bool:
    return type_ is List or (
        isinstance(type_, (_GenericAlias, _SpecialGenericAlias)) and is_sequence(type_)
    )


def is_bare(type_: Any) -> bool:
    args = get_args(type_)
    return (
        args == get_args(List)
        or args == get_args(Sequence)
        or args == get_args(Mapping)
        or args == get_args(Dict)
        or args == get_args(MutableSequence)
    )


def is_mapping_alias(type_: Any) -> bool:
    try:
        return type_.__class__ is _GenericAlias and issubclass(
            type_.__origin__, Mapping
        )
    except TypeError:
        return False


def is_mapping(type_: Any) -> bool:
    return type_ is Mapping or is_mapping_alias(type_)



"""Top-level package for Svarog."""

__author__ = """Damian Świstowski"""
__email__ = "damian@swistowski.org"
__version__ = "0.2.0"

from .svarog import Svarog

default_builder = Svarog()

forge = default_builder.forge
register_forge = default_builder.register_forge

__all__ = ["forge"]



import re

CAMEL_RE_PRE = re.compile("(.)([A-Z][a-z]+)")
CAMEL_RE_POST = re.compile("([a-z0-9])([A-Z])")


def camel_to_snake(name: str) -> str:
    name = CAMEL_RE_PRE.sub(r"\1_\2", name)
    return CAMEL_RE_POST.sub(r"\1_\2", name).lower()



from typing import Any
from typing import Callable
from typing import Type
from typing import TypeVar

T = TypeVar("T")

Forge = Callable[[Type[T], Any], T]
Handler = Callable[[Any, Any, Forge], T]
NoneType = type(None)
Check = Callable[[Any], bool]


class CannotDispatch(Exception):
    pass



from typing import Any
from typing import Callable
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TypeVar

from svarog.types import Check
from svarog.types import Forge
from svarog.types import Handler

T = TypeVar("T")


def _true(*args, **kwargs):
    return True


class FunctionalDispatch:
    __slots__ = ("_registry", "_default")

    def __init__(self, default: Handler):
        self._registry: Sequence[Tuple[Check, Handler]] = [(_true, default)]
        self._default = default

    def register(self, check: Check) -> Callable[[Handler], Handler]:
        def wrapper(handler: Handler) -> Handler:
            self._registry = ((check, handler), *self._registry)
            return handler

        return wrapper

    def __call__(self, type_: Type[T], data: Any, forge: Forge) -> T:
        for check, local_forge in self._registry:
            if check(type_):
                return local_forge(type_, data, forge)
        raise RuntimeError("Cannot find dispatcher")



from functools import lru_cache
from functools import singledispatch
from typing import Any
from typing import Type
from typing import TypeVar

from ..types import CannotDispatch
from ..types import Forge
from .functional import FunctionalDispatch


T = TypeVar("T")


def sentry(type_: Type[T], data: Any, build: Forge) -> T:
    raise CannotDispatch()


class MultiDispatcher:
    def __init__(self):
        self._single_dispatch = singledispatch(CannotDispatch)
        self._functional_dispatch: FunctionalDispatch = FunctionalDispatch(sentry)
        self.dispatch = lru_cache()(self._dispatch)

    def register_cls(self, type_, handler):
        self._single_dispatch.register(type_, handler)
        self.dispatch.cache_clear()

    def register_func(self, func, handler):
        self._functional_dispatch.register(func)(handler)
        self.dispatch.cache_clear()

    def _dispatch(self, type_):
        try:
            dispatch = self._single_dispatch.dispatch(type_)
        except AttributeError:
            dispatch = CannotDispatch

        if dispatch == CannotDispatch:
            return self._functional_dispatch
        return dispatch



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['svarog', 'svarog.dispatchers']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'svarog',
    'version': '0.2.0',
    'description': 'Svarog allow to create object from non typed data',
    'long_description': '======\nSvarog\n======\n\n\n.. image:: https://img.shields.io/pypi/v/svarog.svg\n        :target: https://pypi.python.org/pypi/svarog\n\n.. image:: https://github.com/dswistowski/svarog/actions/workflows/tests.yml/badge.svg\n        :target: https://github.com/dswistowski/svarog/actions/workflows/tests.yml\n\n.. image:: https://readthedocs.org/projects/svarog/badge/?version=latest\n        :target: https://svarog.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nSvarog allow to create object from non typed data. All it need is annotated `__init__` method:\n\n\n>>> from svarog import forge\n... class A:\n...     def __init__(self, a: int, b: str):\n...       self._a = a\n...       self._b = b\n...    def __repr__(self):\n...        return f\'A(a={self._a}, b="{self._b}")\'\n>>> forge(A, {"a": 1, "b": "3"})\nA(a=1, b="3")\n\n\nMore complicated types as `Sequence`, `Mapping`, `Optional` are possible\n\n>>> class A:\n...     def __init__(self, b: Sequence[int]):\n...         self._b = b\n...     def __repr__(self):\n...         return f\'A(b={self._b})\'\n>>> forge(A, {"b": "3213"})\nA(b=[3, 2, 1, 3])\n\nYou can use forward refs:\n\n>>> class WithRef:\n...    def __init__(self, child: Optional[\'WithRef\']):\n...        self._child = child\n...    def __repr__(self):\n...        return f"WithRef({self._child!r})"\n>>> forge(WithRef(WithRef(WithRef())))\nWithRef(WithRef(WithRef(None)))\n\n\nObjects are forged recursively:\n\n\n>>> @dataclass\n... class A:\n...     b: \'B\'\n...     c: \'C\'\n... @dataclass\n... class B:\n...     number: int\n... @dataclass\n... class C:\n...     string: str\n>>> forge(A, {\'b\': {\'number\': 42}, \'c\': {\'string\': \'the-string\'}})\nA(b=B(number=42), c=C(string=\'the-string\'))\n\n\nYou can register own forge for your classes:\n\n>>> class FooType(Enum):\n...     LOREM = "lorem"\n...     IPSUM = "ipsum"\n...\n... class FooParams:\n...     types: ClassVar[Mapping[FooType, "FooParams"]] = {}\n...     def __init_subclass__(cls, type: FooType):\n...        cls.types[type] = cls\n...\n...    @classmethod\n...    def for_type(cls, type):\n...        return cls.types[type]\n...\n... @dataclass\n... class LoremFooParams(FooParams, type=FooType.LOREM):\n...     lorem: str\n...\n... @dataclass\n... class IpsumFooParams(FooParams, type=FooType.IPSUM):\n...     ipsum: int\n...\n... @dataclass\n... class Foo:\n...     type: FooType\n...     params: FooParams\n...\n...     @classmethod\n...     def forge(cls, _, data, forge):\n...         foo_type = forge(FooType, data["type"])\n...         return Foo(\n...             type=forge(FooType, foo_type),\n...             params=forge(FooParams.for_type(foo_type), data["params"])\n...         )\n...\n>>> register_forge(Foo, Foo.forge)\n>>> forge(Foo, {"type": "lorem", "params": {"lorem": "foo-bar"}})\nFoo(type=<FooType.LOREM: \'lorem\'>, params=LoremFooParams(lorem=\'foo-bar\'))\n\n>>> forge(Foo, {"type": "ipsum", "params": {"ipsum": 42}})\nFoo(type=<FooType.IPSUM: \'ipsum\'>, params=IpsumFooParams(ipsum=42))\n\n\nSupport for CamelCase to snake_case convertion:\n\n>>> class Snake:\n...     lorem_ipsum: int\n>>> forge = Svarog(snake_case=True).forge\n>>> forge(Snake, {"LoremIpsum": 42})\nSnake(lorem_ipsum=42)\n\n* Free software: MIT license\n* Documentation: https://svarog.readthedocs.io.\n\n\nFeatures\n--------\n\n* Converts unstructured data into structured recursively\n\n  * Works with `dataclasses`\n  * Works with `Sequence`, `Mapping`, `Optional`\n  * Special conventers for types can be registered with\n\nCredits\n-------\n\nSome parts of this code, and concept borrowed from cattrs_ project\n\n.. _Cattrs: https://github.com/Tinche/cattrs\n\nThis package was created with Cookiecutter_ and the `audreyr/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n',
    'author': 'Damian Świstowski',
    'author_email': 'damian@swistowski.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/dswistowski/svarog/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any
from typing import Optional
from typing import Tuple

try:
    from typing import ForwardRef
except ImportError:
    from typing import _ForwardRef as ForwardRef  # type: ignore

try:
    from typing import _SpecialGenericAlias  # type: ignore
except ImportError:

    class _SpecialGenericAlias:  # type: ignore
        ...


try:
    from typing import get_args
except ImportError:

    def get_args(tp: Any) -> Tuple[Any, ...]:
        return tp.__args__


try:
    from typing import get_origin
except ImportError:

    def get_origin(tp: Any) -> Optional[Any]:
        return tp.__origin__


__all__ = ["ForwardRef", "_SpecialGenericAlias", "get_args", "get_origin"]
